/* 
 * File:   EEPROM.h
 * Author: mhenne
 *
 * Created on April 21, 2015, 7:38 PM
 */

#ifndef EEPROM_H
#define	EEPROM_H

#ifdef	__cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include "configbitsrev2014vC-int.h"
#include "SDlib.h"
#include <xc.h>

/*DECLARATIONS*/
void EEPROM_init(void);
unsigned int EEPROM(unsigned int dataOUT);
void EEPROM_wait(void);
void EEPROM_write(unsigned int data_in, unsigned int addr);
unsigned int EEPROM_read(unsigned int addr);
unsigned int EEPROM_empty_addr(void);
void EEPROM_clear_MEM(void);
void EEPROM_clear_TIME(void);
void EEPROM_clear_current(void);
void EEPROM_clear_all(void);
void EEPROM_new_value(unsigned int num, unsigned int ts);
void EEPROM_read_all(int* array);

/*DEFINITIONS*/
#define CS_eeprom           LATBbits.LATB8
#define IF_eeprom           IFS1bits.SPI4RXIF
#define BUF_eeprom          SPI4BUF

//constants
static const unsigned int CLEAR = 4294967295;
static const int TIME = 0;
static const int MEM_current = 4;
static const int TS_current = 8;
static const int MEM1 = 12; //1st memory location, followed by its timestamp
static const int MEM_total = 10; //total number of glucose readings stored
static const int TS_last = 88;  //timestamp location of last value


/*FUNCTION DEFINITIONS*/
void EEPROM_init(void)
{
    //TS_last = MEM_total*8 + MEM1 - 4;
    unsigned int rData;
    IEC0CLR=0x03800000;	//disable all interrupts
    SPI4CON = 0; // Stops and resets the SPI1.
    rData=SPI4BUF; // clears the receive buffer
    SPI4BRG=4; // needs to be 4 so SPI clock will run at 2MHz (fPB/(2(SPI4BRG+1))=2)
    SPI4CONbits.CKE=1;// changed from 1 to 0?
    SPI4CONbits.CKP=0;
    SPI4CONbits.SMP=1;
    SPI4CONbits.MSTEN=1; // make sure the SPI is in master mode
    SPI4STATbits.SPIROV=0; // clear the overflow bit
    SPI4CONbits.ON=1; // need to turn it on before initializing the EEPROM otherwise things get	reset

    TRISBbits.TRISB8=0; // !CS line to be an output
}

unsigned int EEPROM(unsigned int dataOUT)
{
    IF_eeprom=0; //	clear the interrupt flag
    BUF_eeprom=dataOUT;	// loads the transmit buffer with new data

    while(!IF_eeprom); // waits for all the data to get shifted out by polling the interrupt flag
    unsigned int newData=BUF_eeprom; // get the new data and clear the buffer

    return newData; // return the new data
}

void EEPROM_wait(void)
{
    unsigned int wait = 3;
    CS_eeprom = 1;
    CS_eeprom = 0;

    EEPROM(0x05);
    while (wait == 3){
    wait = EEPROM(0x00);
    }
    CS_eeprom = 1;
}

void EEPROM_write(unsigned int data_in, unsigned int addr)
{
    unsigned short data3, data2, data1, data0, addr1, addr0;

    data3 = data_in >> 24;
    data2 = data_in >> 16;
    data1 = data_in >> 8;
    data0 = data_in;

    addr1 = addr >> 8;
    addr0 = addr;

    CS_eeprom = 0;

    EEPROM(0x06); //write enable command
    CS_eeprom = 1;
    CS_eeprom = 0;
    EEPROM(0x02); //write command
    EEPROM(addr1); //address1
    EEPROM(addr0); //address2
    EEPROM(data3); //data
    EEPROM(data2); //data
    EEPROM(data1); //data
    EEPROM(data0); //data
    EEPROM_wait();

    CS_eeprom = 1;

}

unsigned int EEPROM_read(unsigned int addr)
{
    unsigned int data_out;
    unsigned short dataout3, dataout2, dataout1, dataout0, addr1, addr0;

    addr1 = addr >> 8;
    addr0 = addr;

    CS_eeprom = 0;

    EEPROM(0x03); //read command
    EEPROM(addr1); //address1
    EEPROM(addr0); //address2
    dataout3 = EEPROM(0x00); //send something to recieve something
    dataout2 = EEPROM(0x00); //send something to recieve something
    dataout1 = EEPROM(0x00); //send something to recieve something
    dataout0 = EEPROM(0x00); //send something to recieve something

    CS_eeprom = 1;

    data_out = (dataout3 << 24) | (dataout2 << 16) | (dataout1 << 8) | dataout0;

    return(data_out);
}

unsigned int EEPROM_empty_addr(void)
{
//Gives the number of readings saved in the EEPROM
    unsigned int data, count = MEM1;
    while(data != CLEAR){
        data = EEPROM_read(count);
        count = count + 8;
    }
    count = count - 8;
    return(count);
}

void EEPROM_clear_MEM(void)
{
    unsigned int n;
    for(n = MEM1; n <(TS_last+1); n=n+4){
        EEPROM_write(CLEAR,n);
    }
}

/*
 * Function Name:
 * Description:
 * Input:
 * Outputs:
 */
void EEPROM_clear_TIME(void)
{
    EEPROM_write(CLEAR,TIME);
}

void EEPROM_clear_current(void)
{
    EEPROM_write(CLEAR,MEM_current);
    EEPROM_write(CLEAR,TS_current);
}

void EEPROM_clear_all(void)
{
    EEPROM_clear_MEM();
    EEPROM_clear_TIME();
    EEPROM_clear_current();
}

void EEPROM_new_value(unsigned int num, unsigned int ts)
{
    unsigned int addr;

    EEPROM_write(num, MEM_current);
    EEPROM_write(ts,TS_current);
    addr = EEPROM_empty_addr();
    EEPROM_write(num,addr);
    EEPROM_write(ts,addr+4);
}

void EEPROM_read_all(int* array)
{
    int n,i=0;

    for (n=MEM1; n<=TS_last; n=n+4){
        array[i] = EEPROM_read(n);
        i++;
    }
}



#ifdef	__cplusplus
}
#endif

#endif	/* EEPROM_H */

