//
#include "ble3.h"
#include "../ADHEADER.h"
#include "../Button.h"
#include "../Timer.h"
#include <xc.h>
#include <sys/attribs.h>
#include "../SDlib.h"
#include "../lcd_main.h"
#include "../EEPROM.h"
#include "../sleepMode.h"
#include <math.h>
//#include "send_timestamp.h"

#define NELEMS(x)  (sizeof(x) / sizeof((x)[0]))

//constants
static const int timer_test_strip_insertion = 30;
static const int timer_add_blood = 60;
static const int timer_ble_connection = 30;
static const int timer_ble_transfer = 10;


//function declarations
void low_power_init(void);
void start_FSM(void);
void interrupt_init(void);
int fsm(void);
void main_init(void);
int timeout(int time, int *counter, int *countdown);
char* stringWithNumber(char *s1, int num);
float ADC_FSM(void);
float VOLTAGE_TO_BLOOD_SUGAR(float voltage);
int BLE_FSM(int data[32], int numEntries);
int BLE_FSM_Connect(int data);
int BLE_FSM_Transfer(int data);
int BLE_FSM_Receive_Time(void);

/***	External Interrupt 1
**
**	Parameters:
**		_EXTERNAL_2_VECTOR interrupt (INT2)
**
**	Return Value:
**		none
**
**	Errors:
**		none
**
**	Description:
**		This function is fired when the left button is clicked
 *              Used to wake up and show the clock and/or temperture
*/
void __ISR(_EXTERNAL_2_VECTOR, IPL7AUTO) button1(void)
{
    if(IFS0bits.INT2IF == 1) {
        IFS0bits.INT2IF = 0;    //reset interrupt flag
        delay_ms(100);          //button debounce
    } else {
        //do nothing
    }
}

/***	External Interrupt 2
**
**	Parameters:
**		_EXTERNAL_4_VECTOR interrupt (INT4)
**
**	Return Value:
**		none
**
**	Errors:
**		none
**
**	Description:
**		This function is fired when the right button is clicked
 *              Used to wake up and start the read blood sugar routine
*/
void __ISR(_EXTERNAL_4_VECTOR, IPL7AUTO) button2(void)
{
    if(IFS0bits.INT4IF == 1) {
        IFS0bits.INT4IF = 0;    //reset interrupt flag
        delay_ms(100);          //button debounce

        start_FSM();
    } else {
        //do nothing
    }
}


/***	Main
**
**	Parameters:
**		none
**
**	Return Value:
**		none
**
**	Errors:
**		none
**
**	Description:
**		Main function that starts running the code on a reset.
*/
void main(void)
{
    DDPCONbits.JTAGEN=0; // turns off the JTAG

    interrupt_init();
    //timer_clk_start();  //begin our clock counting every second
    sleep_mode_setup(); 
    
    while(1) {
        sleep();  // always sleep when we are not running code
    };

}

/***	start_FSM
**
**	Parameters:
**		none
**
**	Return Value:
**		none
**
**	Errors:
**		none
**
**	Description:
**		Transistions from sleep to woken up and runs init code necassary to set up the system
*/
void start_FSM(void) {
    IEC0bits.INT2IE = 0;    //disable interrupt routine
    IEC0bits.INT4IE = 0;    //disable interrupt routine
    //__builtin_disable_interrupts();   //disable interrupts
    main_init();

    int running;
    running = fsm();

    /*running is either going to be 1 after a full cycle of completion of the code or 0 when it is entering
     *sleep mode.
     */
    while(running != 0) {
        if(running == 2) {
            running = fsm();
        }
    }

    OledClearBuffer();
    OledUpdate();
    interrupt_init();
}

/***	interrupt_init
**
**	Parameters:
**		none
**
**	Return Value:
**		none
**
**	Errors:
**		none
**
**	Description:
**		Initializes all of the init code for interrupts, called before
 *              we go to sleep so that we can be woken up.
*/
void interrupt_init(void) {
    INTCONbits.MVEC = 1;
    button_1_init();
    button_2_init();

    __builtin_enable_interrupts();   //enable interrupts
}

/***	fsm
**
**	Parameters:
**		none
**
**	Return Value:
**		int 1: reached end of code
 *                  0: ready to turn off and go to sleep
**
**	Errors:
**		none
**
**	Description:
**		Finite State Machine for the blood sugar reading sequence.
*/
int fsm(void) {
    
    int button_value;

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Insert Strip");

    OledSetCursor(0,1);
    OledPutString("Press button when ready");

    int counter, countdown, shutdown, blink, position;
    position = 0;
    shutdown = 1;
    counter = 0;
    countdown = 3;
    blink = 0;
    button_value = button_pressed();
    timer_start();
    while(button_value == 1) {
        button_value = button_pressed();
        shutdown = timeout(timer_test_strip_insertion,&counter,&countdown);
        if(shutdown == 0){
            //actual shutdown code here
            return 0;
        }
    }

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Please Add Blood to the strip");

    float blood_sugar;
    blood_sugar = ADC_FSM();
    while(blood_sugar == 0.0f) {
        blood_sugar = ADC_FSM();
        if(blood_sugar == 1.0f) {
            OledClearBuffer();
            OledSetCursor(0,0);
            OledPutString("Bad Reading");
            delay_ms(2000);
            return 2;  //restart from beginning
        }
        shutdown = timeout(timer_add_blood,&counter,&countdown);
        if(shutdown == 0){
            //actual shutdown code here
            return 0;
        }
    }

    write_blood_glucose_reading(blood_sugar);
    delay_ms(4000);


    /*EEPROM_init();
    EEPROM_new_value((int)blood_sugar, 1234); //INSERT TIMESTAMP in 2nd argument

    int  array[MEM_total*2];
    EEPROM_read_all(array);

     */

    int array[64];
    array[0] = (int)blood_sugar;
    array[1] = 1111111111;

    int n;
    n = BLE_FSM(array,1);
    if(n == 0) {
        return 0;
    }

    write_blood_glucose_reading(blood_sugar);
    delay_ms(5000);


    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("end of test");

    return 1;
}

/***	main_init
**
**	Parameters:
**		none
**
**	Return Value:
**		none
**
**	Errors:
**		none
**
**	Description:
**		Initializes the OLED screen and ADC for the blood sugar reading
*/
void main_init(void) {
    DeviceInit();
    ADC_init();
    NVMCONbits.WREN = 1;  //enable low voltage detect circuit
    if(NVMCONbits.LVDERR == 1) {  //low voltage?
        OledClearBuffer();
        OledSetCursor(0,0);
        OledPutString("low battery");
    }
    NVMCONbits.WREN = 0;  //disable LVD circuit
}

/***	ADC_FSM
**
**	Parameters:
**		none
**
**	Return Value:
**		returns the blood sugar reading in a float or 0.0 if no reading
 *              taken
**
**	Errors:
**		none
**
**	Description:
**		Waits until blood is applied to the test strip, then converts
 *              the ADC reading of a voltage from 0-3.3V to an actual blood
 *              sugar.  Temperature is read in this sequence as it affects the
 *              blood sugar reading.
*/
float ADC_FSM(void) {
    float temp,test_strip,blood_sugar;
    float total,samples,avg;
    samples = 10.0f;
    temp = ADC_read_temperature();
    test_strip = ADC_read_test_strip();
    if(test_strip > 0.95){
        temp = ADC_read_temperature();
        delay_ms(1100);
        int i;
        total = 0.0f;
        total = ADC_read_test_strip() - 0.9f;
        if(total < .2) {  //bad reading
            return 1.0;
        }
        for(i = 1; i < 10; i++) {
            total += ADC_read_test_strip() - 0.9f;
        }
        avg = total / samples;
        //avg = (ADC_read_test_strip() - .9) *100;
        //avg -= ( temp - 68.0f) / 100.0f;   //accuracy of this equation would be found with lab testing procedures
        blood_sugar = VOLTAGE_TO_BLOOD_SUGAR(avg);

        return blood_sugar;
    }
    return 0.0;
}

float VOLTAGE_TO_BLOOD_SUGAR(float voltage) {
    float blood_sugar;

    blood_sugar = 262.25 * log(voltage) + 407.24;

    if(blood_sugar < 100 && blood_sugar > 60) {
        blood_sugar = blood_sugar - 20;
    }

    return blood_sugar;
}

/***	BLE_FSM
**
**	Parameters:
**		int data: data to send
**
**	Return Value:
**		returns 0 if BLE timesout (connecting or transfer)
**              returns 1 successful completion of BLE transferring
 * 
**	Errors:
**		none
**
**	Description:
**		After the reading has successfully completed, the bluetooth
 *              FSM begins.
*/
int BLE_FSM(int data[64], int numEntries) {
    int cont, index, numOfPoints;

    ble_setup();

    cont = BLE_FSM_Connect(999);
    if(cont == 0) {
        return 0;
    }

    numOfPoints = NELEMS(data);
    for(index = 0; index < numEntries*2; index++) {
        if(data[index] < 999999999)
        {
            cont = BLE_FSM_Transfer(data[index]);
        }
        else
        {
            if ((int)data[index] % 100 == 0)
                data[index] = data[index] + 11;

            cont =  BLE_FSM_Transfer((int)data[index]/100);
            cont = BLE_FSM_Transfer((int)data[index] % 100);
        }
        if(cont == 0) {
            return 0;
        }
    }

//    cont = BLE_FSM_Receive_Time();
//    if(cont == 0) {
//        return 0;
//    }
    
    return 1;
}

/***	BLE_FSM_Connect
**
**	Parameters:
**		int data: data to send
**
**	Return Value:
**		returns 0 if BLE times out (connecting)
**              returns 1 successful completion of BLE transferring
 *
**	Errors:
**		none
**
**	Description:
**		Attempts to connect bluetooth chip to mobile app
*/
int BLE_FSM_Connect(int data) {
    int counter, countdown, shutdown;
    shutdown = 1;
    counter = 0;
    countdown = 3;

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Open app & press connect");

    delay_ms(4000);

    bool connected = false;
    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Connecting...");
    delay_ms(1000);
    while(connected == false) {
        connected = ble_connect(data);

        shutdown = timeout(timer_ble_connection,&counter,&countdown);
        if(shutdown == 0){
            OledClearBuffer();
            OledSetCursor(0,0);
            OledPutString("Couldn't Connect");
            delay_ms(2000);

            OledClearBuffer();
            OledSetCursor(0,0);
            OledPutString("Storing for later");
            delay_ms(2000);
            return 0;
        }
    }

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Connected");
    delay_ms(1000);

    return 1;
}

/***	BLE_FSM_Transfer
**
**	Parameters:
**		int data: data to send
**
**	Return Value:
**		returns 0 if BLE times out (unsuccessful transfer)
**              returns 1 successful completion of BLE transferring
 *
**	Errors:
**		none
**
**	Description:
**		Attempts to transfer data to mobile app
*/
int BLE_FSM_Transfer(int data){
    int counter, countdown, shutdown, button_value;
    bool complete = false;
    shutdown = 1;
    counter = 0;
    countdown = 3;
//
//    OledClearBuffer();
//    OledSetCursor(0,0);
//    OledPutString("Press button to transfer");
//
//    button_value = button_pressed();
//    while(button_value == 1) {
//        button_value = button_pressed();
//        shutdown = timeout(30,&counter,&countdown);
//        if(shutdown == 0){
//            //actual shutdown code here
//            return 0;
//        }
//    }

    shutdown = 1;
    counter = 0;
    countdown = 3;

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Transferring...");
    delay_ms(2500);
    while(complete == false) {
        complete = ble_transfer(data);

        shutdown = timeout(timer_ble_transfer,&counter,&countdown);
        if(shutdown == 0){
            OledClearBuffer();
            OledSetCursor(0,0);
            OledPutString("Couldn't Tranfer");
            delay_ms(2000);
            OledClearBuffer();
            OledSetCursor(0,0);
            OledPutString("Storing for later");
            delay_ms(2000);
            return 0;
        }
    }

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Transferred");
    EEPROM_clear_MEM();
    delay_ms(1000);

    return 1;
}

int BLE_FSM_Receive_Time(void){
    int counter, countdown, shutdown;
    bool complete = false;

    shutdown = 1;
    counter = 0;
    countdown = 3;

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("Receiving...");
    delay_ms(500);
    int new_timestamp;
    new_timestamp = ble_receive();

    OledClearBuffer();
    OledSetCursor(0,0);
    OledPutString("received");
    delay_ms(3000);

    return new_timestamp;
}


